# Copyright 2021 The Cirq Developers
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=missing-function-docstring
# pylint: disable=missing-return-doc
# mypy: disable-error-code=method-assign
from __future__ import annotations

import itertools
import re
from unittest.mock import MagicMock

import cirq
import numpy as np
import pandas as pd
import pytest

from supermarq.qcvv import XEB, XEBResults


def test_xeb_init() -> None:
    experiment = XEB(num_circuits=10, cycle_depths=[1, 3, 5])
    assert experiment.num_qubits == 2
    assert experiment.two_qubit_gate == cirq.CZ
    assert experiment.single_qubit_gate_set == [
        cirq.PhasedXZGate(
            z_exponent=z,
            x_exponent=0.5,
            axis_phase_exponent=a,
        )
        for a, z in itertools.product(np.linspace(start=0, stop=7 / 4, num=8), repeat=2)
    ]

    experiment = XEB(two_qubit_gate=cirq.CX, num_circuits=10, cycle_depths=[1, 3, 5])
    assert experiment.num_qubits == 2
    assert experiment.two_qubit_gate == cirq.CX
    assert experiment.single_qubit_gate_set == [
        cirq.PhasedXZGate(
            z_exponent=z,
            x_exponent=0.5,
            axis_phase_exponent=a,
        )
        for a, z in itertools.product(np.linspace(start=0, stop=7 / 4, num=8), repeat=2)
    ]

    experiment = XEB(single_qubit_gate_set=[cirq.X], num_circuits=10, cycle_depths=[1, 3, 5])
    assert experiment.num_qubits == 2
    assert experiment.two_qubit_gate == cirq.CZ
    assert experiment.single_qubit_gate_set == [cirq.X]


@pytest.fixture
def xeb_experiment() -> XEB:
    return XEB(
        single_qubit_gate_set=[cirq.X, cirq.Y, cirq.Z], num_circuits=10, cycle_depths=[1, 3, 5]
    )


def test_build_xeb_circuit(xeb_experiment: XEB) -> None:

    xeb_experiment._rng = (rng := MagicMock())
    rng.choice.side_effect = [
        np.array([[cirq.X, cirq.Y], [cirq.Z, cirq.Y], [cirq.Y, cirq.Z]]),
        np.array([[cirq.X, cirq.Z], [cirq.X, cirq.X], [cirq.Y, cirq.Y]]),
    ]
    circuits = xeb_experiment._build_circuits(num_circuits=2, cycle_depths=[2])

    assert len(circuits) == 2

    qbs = xeb_experiment.qubits
    cirq.testing.assert_same_circuits(
        circuits[0].circuit,
        cirq.Circuit(
            [
                cirq.X(qbs[0]),
                cirq.Y(qbs[1]),
                cirq.TaggedOperation(cirq.CZ(*qbs), "no_compile"),
                cirq.Z(qbs[0]),
                cirq.Y(qbs[1]),
                cirq.TaggedOperation(cirq.CZ(*qbs), "no_compile"),
                cirq.Y(qbs[0]),
                cirq.Z(qbs[1]),
                cirq.measure(qbs),
            ]
        ),
    )
    assert circuits[0].data == {
        "circuit_depth": 5,
        "cycle_depth": 2,
        "two_qubit_gate": "CZ",
        "exact_00": 1.0,
        "exact_01": 0.0,
        "exact_10": 0.0,
        "exact_11": 0.0,
    }
    cirq.testing.assert_same_circuits(
        circuits[1].circuit,
        cirq.Circuit(
            [
                cirq.X(qbs[0]),
                cirq.Z(qbs[1]),
                cirq.TaggedOperation(cirq.CZ(*qbs), "no_compile"),
                cirq.X(qbs[0]),
                cirq.X(qbs[1]),
                cirq.TaggedOperation(cirq.CZ(*qbs), "no_compile"),
                cirq.Y(qbs[0]),
                cirq.Y(qbs[1]),
                cirq.measure(qbs),
            ]
        ),
    )
    assert circuits[1].data == {
        "circuit_depth": 5,
        "cycle_depth": 2,
        "two_qubit_gate": "CZ",
        "exact_00": 0.0,
        "exact_01": 0.0,
        "exact_10": 1.0,
        "exact_11": 0.0,
    }


def test_xeb_analyse_results(xeb_experiment: XEB) -> None:
    results = XEBResults(target="example", experiment=xeb_experiment)

    results.data = pd.DataFrame(
        [
            {
                "circuit_index": 0,
                "cycle_depth": 1,
                "circuit_depth": 3,
                "00": 1.0,
                "01": 0.0,
                "10": 0.0,
                "11": 0.0,
                "exact_00": 1.0,
                "exact_01": 0.0,
                "exact_10": 0.0,
                "exact_11": 0.0,
            },
            {
                "circuit_index": 1,
                "cycle_depth": 1,
                "circuit_depth": 3,
                "00": 1.0,
                "01": 0.0,
                "10": 0.0,
                "11": 0.0,
                "exact_00": 0.5,
                "exact_01": 0.5,
                "exact_10": 0.0,
                "exact_11": 0.0,
            },
            {
                "circuit_index": 0,
                "cycle_depth": 5,
                "circuit_depth": 11,
                "00": 0.0,
                "01": 1.0,
                "10": 0.0,
                "11": 0.0,
                "exact_00": 0.0,
                "exact_01": 0.75,
                "exact_10": 0.25,
                "exact_11": 0.0,
            },
            {
                "circuit_index": 1,
                "cycle_depth": 5,
                "circuit_depth": 11,
                "00": 0.0,
                "01": 0.0,
                "10": 1.0,
                "11": 0.0,
                "exact_00": 0.0,
                "exact_01": 0.5,
                "exact_10": 0.25,
                "exact_11": 0.25,
            },
            {
                "circuit_index": 0,
                "cycle_depth": 10,
                "circuit_depth": 21,
                "00": 0.0,
                "01": 0.0,
                "10": 0.5,
                "11": 0.5,
                "exact_00": 0.2,
                "exact_01": 0.3,
                "exact_10": 0.25,
                "exact_11": 0.25,
            },
            {
                "circuit_index": 1,
                "cycle_depth": 10,
                "circuit_depth": 21,
                "00": 0.0,
                "01": 0.0,
                "10": 0.5,
                "11": 0.5,
                "exact_00": 0.1,
                "exact_01": 0.1,
                "exact_10": 0.4,
                "exact_11": 0.4,
            },
        ]
    )
    results.analyze()
    np.testing.assert_allclose(
        results.data["sum_p(x)p^(x)"].values, [1.0, 0.5, 0.75, 0.25, 0.25, 0.4]
    )
    np.testing.assert_allclose(
        results.data["sum_p(x)p(x)"].values, [1.0, 0.5, 0.625, 0.375, 0.255, 0.34]
    )

    # Calculated by hand
    assert results.cycle_fidelity_estimate == pytest.approx(1.0613025)
    assert results.cycle_fidelity_estimate_std == pytest.approx(0.0597633930)

    results.plot_results()
    results.plot_speckle()


def test_results_no_data() -> None:
    results = XEBResults(target="example", experiment=MagicMock(), data=None)
    with pytest.raises(RuntimeError, match="No data stored. Cannot perform analysis."):
        results._analyze()

    with pytest.raises(RuntimeError, match="No data stored. Cannot plot results."):
        results.plot_results()

    with pytest.raises(
        RuntimeError, match="No stored dataframe of circuit fidelities. Something has gone wrong."
    ):
        results.data = pd.DataFrame()
        results.plot_results()


def test_results_not_analyzed() -> None:
    results = XEBResults(target="example", experiment=MagicMock(), data=None)
    for attr in ["cycle_fidelity_estimate", "cycle_fidelity_estimate_std"]:
        with pytest.raises(
            RuntimeError,
            match=re.escape("Value has not yet been estimated. Please run `.analyze()` method."),
        ):
            getattr(results, attr)



# Copyright 2021 The Cirq Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=missing-function-docstring
# pylint: disable=missing-return-doc
# mypy: disable-error-code=method-assign

from __future__ import annotations

import re
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from unittest.mock import MagicMock, call, patch

import cirq
import cirq_superstaq as css
import numpy as np
import pandas as pd
import pytest

from supermarq.qcvv.base_experiment import QCVVExperiment, QCVVResults, Sample

mock_plot = MagicMock()
mock_print = MagicMock()


@dataclass
class ExampleResults(QCVVResults):
    """Example results class for testing"""

    _example_final_result: float | None = None

    def _analyze(self) -> None:
        self._example_final_result = 3.142

    def plot_results(self) -> None:
        mock_plot()

    def print_results(self) -> None:
        mock_print("This is a test")

    @property
    def example_final_result(self) -> float:
        if self._example_final_result is None:
            raise self._not_analyzed
        return self._example_final_result


class ExampleExperiment(QCVVExperiment[ExampleResults]):
    """Example experiment class for testing"""

    def _build_circuits(self, num_circuits: int, cycle_depths: Iterable[int]) -> Sequence[Sample]:
        return [
            Sample(
                circuit=MagicMock(spec=cirq.Circuit), data={"num": k, "depth": d}, circuit_index=k
            )
            for k in range(num_circuits)
            for d in cycle_depths
        ]


@pytest.fixture
def abc_experiment() -> ExampleExperiment:
    with patch("supermarq.qcvv.base_experiment.QCVVExperiment._validate_circuits"):
        return ExampleExperiment(
            num_qubits=2,
            num_circuits=10,
            cycle_depths=[1, 3, 5],
            random_seed=42,
            results_cls=ExampleResults,
            service_details="Some other details",
        )


@pytest.fixture
def sample_circuits() -> list[Sample]:
    qubits = cirq.LineQubit.range(2)
    return [
        Sample(
            circuit=cirq.Circuit(cirq.CZ(*qubits), cirq.CZ(*qubits), cirq.measure(*qubits)),
            data={"circuit": 1},
            circuit_index=1,
        ),
        Sample(
            circuit=cirq.Circuit(cirq.CX(*qubits), cirq.measure(*qubits)),
            data={"circuit": 2},
            circuit_index=2,
        ),
    ]


def test_qcvv_experiment_init(
    abc_experiment: ExampleExperiment,
) -> None:
    assert abc_experiment.num_qubits == 2
    assert abc_experiment.num_circuits == 10
    assert abc_experiment.cycle_depths == [1, 3, 5]
    assert abc_experiment._results_cls == ExampleResults
    assert abc_experiment._service_kwargs == {"service_details": "Some other details"}
    assert len(abc_experiment.samples) == 30
    assert isinstance(abc_experiment._rng, np.random.Generator)


def test_results_init(
    abc_experiment: ExampleExperiment,
) -> None:
    results = ExampleResults(
        target="target", experiment=abc_experiment, job=MagicMock(spec=css.Job)
    )
    assert results.target == "target"
    assert results.samples == abc_experiment.samples
    assert results.num_circuits == 10
    assert results.num_qubits == 2


def test_experiment_init_with_bad_layers() -> None:
    with pytest.raises(
        ValueError, match="The `cycle_depths` iterator can only include positive values."
    ):
        ExampleExperiment(
            num_qubits=2,
            num_circuits=10,
            cycle_depths=[0],
            random_seed=42,
            results_cls=ExampleResults,
            service_details="Some other details",
        )


def test_results_not_analyzed(abc_experiment: ExampleExperiment) -> None:
    results = ExampleResults(
        target="target", experiment=abc_experiment, job=MagicMock(spec=css.Job)
    )
    with pytest.raises(
        RuntimeError,
        match=re.escape("Value has not yet been estimated. Please run `.analyze()` method."),
    ):
        _ = results.example_final_result


def test_results_job_still_running(abc_experiment: ExampleExperiment) -> None:
    results = ExampleResults(
        target="target", experiment=abc_experiment, job=MagicMock(spec=css.Job)
    )
    results.job.status.return_value = "Pending"  # type: ignore[union-attr]
    with pytest.warns(
        Warning,
        match=(
            "Experiment data is not yet ready to analyse. This is likely because "
            "the Superstaq job has not yet been completed. Either wait and try again "
            "later, or interrogate the `.job` attribute."
        ),
    ):
        results.analyze()


def test_results_job_no_data(abc_experiment: ExampleExperiment) -> None:
    results = ExampleResults(
        target="target",
        experiment=abc_experiment,
    )
    with pytest.raises(
        RuntimeError,
        match=(
            "No data available and no Superstaq job to use to collect data. Please manually add "
            "results data in order to perform analysis"
        ),
    ):
        results.analyze()


def test_results_analyze(abc_experiment: ExampleExperiment) -> None:
    results = ExampleResults(
        target="target", experiment=abc_experiment, data=MagicMock(spec=pd.DataFrame)
    )

    results.analyze(plot_results=True, print_results=True)
    assert results.example_final_result == 3.142
    mock_plot.assert_called_once()
    mock_print.assert_called_once_with("This is a test")


def test_results_ready(abc_experiment: ExampleExperiment) -> None:
    results = ExampleResults(
        target="target", experiment=abc_experiment, data=MagicMock(spec=pd.DataFrame)
    )
    assert results.data_ready


def test_results_ready_from_job(
    abc_experiment: ExampleExperiment, sample_circuits: list[Sample]
) -> None:
    abc_experiment.samples = sample_circuits
    results = ExampleResults(
        target="target", experiment=abc_experiment, job=MagicMock(spec=css.Job)
    )
    results.job.status.return_value = "Done"  # type: ignore[union-attr]
    results.job.counts.return_value = [  # type: ignore[union-attr]
        {
            "00": 20,
            "01": 5,
            "11": 10,
        },
        {
            "00": 30,
            "01": 5,
        },
    ]
    assert results.data_ready
    pd.testing.assert_frame_equal(
        results.data,
        pd.DataFrame(
            [
                {"circuit": 1, "00": 20 / 35, "01": 5 / 35, "10": 0.0, "11": 10 / 35},
                {"circuit": 2, "00": 30 / 35, "01": 5 / 35, "10": 0.0, "11": 0.0},
            ]
        ),
        check_like=True,
    )


def test_run_with_simulator(
    abc_experiment: ExampleExperiment, sample_circuits: list[Sample]
) -> None:
    cirq.measurement_key_name = MagicMock()
    abc_experiment.samples = sample_circuits
    test_sim = MagicMock()
    mock_result = MagicMock()
    mock_result.histogram.return_value = {0: 0, 1: 100, 2: 0, 3: 0}
    test_sim.run.return_value = mock_result

    results = abc_experiment.run_with_simulator(simulator=test_sim, repetitions=100)

    # Test simulator calls
    test_sim.run.assert_has_calls(
        [
            call(sample_circuits[0].circuit, repetitions=100),
            call(sample_circuits[1].circuit, repetitions=100),
        ],
        any_order=True,
    )

    assert results.experiment == abc_experiment
    assert results.target == "local_simulator"

    # Check the data is stored
    pd.testing.assert_frame_equal(
        results.data,
        pd.DataFrame(
            [
                {"circuit_index": 1, "circuit": 1, "00": 0.0, "01": 1.0, "10": 0.0, "11": 0.0},
                {"circuit_index": 2, "circuit": 2, "00": 0.0, "01": 1.0, "10": 0.0, "11": 0.0},
            ]
        ),
    )


def test_run_with_simulator_default_target(
    abc_experiment: ExampleExperiment, sample_circuits: list[Sample]
) -> None:
    cirq.measurement_key_name = MagicMock()
    cirq.Simulator = (target := MagicMock())  # type: ignore [misc]
    abc_experiment.samples = sample_circuits
    mock_result = MagicMock()
    mock_result.histogram.return_value = {0: 0, 1: 100, 2: 0, 3: 0}
    target().run.return_value = mock_result

    results = abc_experiment.run_with_simulator(repetitions=100)

    # Test simulator calls
    target().run.assert_has_calls(
        [
            call(sample_circuits[0].circuit, repetitions=100),
            call(sample_circuits[1].circuit, repetitions=100),
        ],
        any_order=True,
    )

    assert results.experiment == abc_experiment
    assert results.target == "local_simulator"

    # Check the data is stored
    pd.testing.assert_frame_equal(
        results.data,
        pd.DataFrame(
            [
                {"circuit_index": 1, "circuit": 1, "00": 0.0, "01": 1.0, "10": 0.0, "11": 0.0},
                {"circuit_index": 2, "circuit": 2, "00": 0.0, "01": 1.0, "10": 0.0, "11": 0.0},
            ]
        ),
    )


def test_run_on_device(abc_experiment: ExampleExperiment, sample_circuits: list[Sample]) -> None:
    abc_experiment.samples = sample_circuits

    with patch("cirq_superstaq.Service") as mock_service:
        results = abc_experiment.run_on_device(
            target="example_target", repetitions=100, **{"some": "options"}
        )

    mock_service.return_value.create_job.assert_called_once_with(
        [sample_circuits[0].circuit, sample_circuits[1].circuit],
        target="example_target",
        method=None,
        repetitions=100,
        some="options",
    )

    assert results.job == mock_service.return_value.create_job.return_value
    assert results.target == "example_target"
    assert results.experiment == abc_experiment


def test_run_on_device_dry_run(
    abc_experiment: ExampleExperiment, sample_circuits: list[Sample]
) -> None:
    abc_experiment.samples = sample_circuits

    with patch("cirq_superstaq.Service") as mock_service:
        results = abc_experiment.run_on_device(
            target="example_target", repetitions=100, method="dry-run"
        )

    mock_service.return_value.create_job.assert_called_once_with(
        [sample_circuits[0].circuit, sample_circuits[1].circuit],
        target="example_target",
        method="dry-run",
        repetitions=100,
    )
    assert results.job == mock_service.return_value.create_job.return_value
    assert results.target == "example_target"
    assert results.experiment == abc_experiment


def test_interleave_circuit() -> None:
    qubit = cirq.LineQubit(0)
    circuit = cirq.Circuit(*[cirq.X(qubit) for _ in range(4)])

    # With last gate
    interleaved_circuit = QCVVExperiment._interleave_op(circuit, cirq.Z(qubit), include_final=True)
    cirq.testing.assert_same_circuits(
        interleaved_circuit,
        cirq.Circuit(
            cirq.X(qubit),
            cirq.TaggedOperation(cirq.Z(qubit), "no_compile"),
            cirq.X(qubit),
            cirq.TaggedOperation(cirq.Z(qubit), "no_compile"),
            cirq.X(qubit),
            cirq.TaggedOperation(cirq.Z(qubit), "no_compile"),
            cirq.X(qubit),
            cirq.TaggedOperation(cirq.Z(qubit), "no_compile"),
        ),
    )

    # Without last gate
    interleaved_circuit = QCVVExperiment._interleave_op(circuit, cirq.Z(qubit), include_final=False)
    cirq.testing.assert_same_circuits(
        interleaved_circuit,
        cirq.Circuit(
            cirq.X(qubit),
            cirq.TaggedOperation(cirq.Z(qubit), "no_compile"),
            cirq.X(qubit),
            cirq.TaggedOperation(cirq.Z(qubit), "no_compile"),
            cirq.X(qubit),
            cirq.TaggedOperation(cirq.Z(qubit), "no_compile"),
            cirq.X(qubit),
        ),
    )


def test_validate_circuits(
    abc_experiment: ExampleExperiment, sample_circuits: list[Sample]
) -> None:
    # Should't get any errors with the base circuits
    abc_experiment._validate_circuits(sample_circuits)

    # Add a gate so not all measurements are terminal
    sample_circuits[0].circuit += cirq.X(abc_experiment.qubits[0])
    with pytest.raises(
        ValueError, match="QCVV experiment circuits can only contain terminal measurements."
    ):
        abc_experiment._validate_circuits(sample_circuits)

    # Remove measurements
    sample_circuits[0].circuit = sample_circuits[0].circuit[:-2] + cirq.measure(
        abc_experiment.qubits[0]
    )
    with pytest.raises(
        ValueError,
        match="The terminal measurement in QCVV experiment circuits must measure all qubits.",
    ):
        abc_experiment._validate_circuits(sample_circuits)

    # Remove all measurements
    sample_circuits[0].circuit = sample_circuits[0].circuit[:-2]
    with pytest.raises(
        ValueError,
        match="QCVV experiment circuits must contain measurements.",
    ):
        abc_experiment._validate_circuits(sample_circuits)


def test_run_with_callable(
    abc_experiment: ExampleExperiment,
    sample_circuits: list[Sample],
) -> None:
    abc_experiment.samples = sample_circuits
    test_callable = MagicMock()
    test_callable.return_value = {"01": 0.2, "10": 0.7, "11": 0.1}

    results = abc_experiment.run_with_callable(test_callable, some="kwargs")

    test_callable.assert_has_calls(
        [
            call(sample_circuits[0].circuit, some="kwargs"),
            call(sample_circuits[1].circuit, some="kwargs"),
        ]
    )

    assert results.target == "callable"
    assert results.experiment == abc_experiment

    # Check the data is stored
    pd.testing.assert_frame_equal(
        results.data,
        pd.DataFrame(
            [
                {"circuit": 1, "00": 0.0, "01": 0.2, "10": 0.7, "11": 0.1},
                {"circuit": 2, "00": 0.0, "01": 0.2, "10": 0.7, "11": 0.1},
            ]
        ),
        check_like=True,
    )


def test_run_with_callable_mixd_keys(
    abc_experiment: ExampleExperiment,
    sample_circuits: list[Sample],
) -> None:
    abc_experiment.samples = sample_circuits
    test_callable = MagicMock()
    test_callable.return_value = {1: 0.2, "10": 0.7, 3: 0.1}

    results = abc_experiment.run_with_callable(test_callable, some="kwargs")

    test_callable.assert_has_calls(
        [
            call(sample_circuits[0].circuit, some="kwargs"),
            call(sample_circuits[1].circuit, some="kwargs"),
        ]
    )

    assert results.target == "callable"
    assert results.experiment == abc_experiment

    # Check the data is stored
    pd.testing.assert_frame_equal(
        results.data,
        pd.DataFrame(
            [
                {"circuit": 1, "00": 0.0, "01": 0.2, "10": 0.7, "11": 0.1},
                {"circuit": 2, "00": 0.0, "01": 0.2, "10": 0.7, "11": 0.1},
            ]
        ),
        check_like=True,
    )


def test_run_with_callable_bad_bitstring(
    abc_experiment: ExampleExperiment,
    sample_circuits: list[Sample],
) -> None:
    abc_experiment.samples = sample_circuits
    test_callable = MagicMock()
    test_callable.return_value = {"000": 0.0, "01": 0.2, "10": 0.8}

    with pytest.raises(
        ValueError,
        match=("The key contains the wrong number of bits. Got 3 entries " "but expected 2 bits."),
    ):
        abc_experiment.run_with_callable(test_callable, some="kwargs")


def test_run_with_callable_bad_probabilities(
    abc_experiment: ExampleExperiment,
    sample_circuits: list[Sample],
) -> None:
    abc_experiment.samples = sample_circuits
    test_callable = MagicMock()
    test_callable.return_value = {"00": 0.0, "01": 0.2, "10": 0.7, "11": 0.09}

    with pytest.raises(RuntimeError, match="Provided probabilities do not sum to 1.0."):
        abc_experiment.run_with_callable(test_callable, some="kwargs")


def test_results_collect_device_counts(
    abc_experiment: ExampleExperiment, sample_circuits: list[Sample]
) -> None:
    abc_experiment.samples = sample_circuits
    results = ExampleResults(
        target="example_target", experiment=abc_experiment, job=MagicMock(spec=css.Job)
    )
    results.job.counts.return_value = [  # type: ignore[union-attr]
        {
            "00": 20,
            "01": 5,
            "11": 10,
        },
        {
            "00": 30,
            "01": 5,
        },
    ]

    df = results._collect_device_counts()

    pd.testing.assert_frame_equal(
        df,
        pd.DataFrame(
            [
                {"circuit": 1, "00": 20 / 35, "01": 5 / 35, "10": 0.0, "11": 10 / 35},
                {"circuit": 2, "00": 30 / 35, "01": 5 / 35, "10": 0.0, "11": 0.0},
            ]
        ),
        check_like=True,
    )


def test_results_collect_device_counts_no_job() -> None:
    results = ExampleResults(target="example_target", experiment=MagicMock(), job=None)
    with pytest.raises(
        ValueError,
        match=("No Superstaq job associated with these results. Cannot collect device counts."),
    ):
        results._collect_device_counts()


def test_canonicalize_bitstring() -> None:
    assert QCVVExperiment._canonicalize_bitstring("00", 2) == "00"
    assert QCVVExperiment._canonicalize_bitstring(1, 2) == "01"
    assert QCVVExperiment._canonicalize_bitstring(5, 4) == "0101"

    with pytest.raises(ValueError, match="The key must be positive. Instead got -2."):
        QCVVExperiment._canonicalize_bitstring(-2, 4)

    with pytest.raises(
        ValueError,
        match=(
            "The key is too large to be encoded with 4 qubits. Got 72 " "but expected less than 16."
        ),
    ):
        QCVVExperiment._canonicalize_bitstring(72, 4)

    with pytest.raises(
        ValueError,
        match=("The key contains the wrong number of bits. Got 5 entries " "but expected 4 bits."),
    ):
        QCVVExperiment._canonicalize_bitstring("01010", 4)

    with pytest.raises(ValueError, match="All entries in the bitstring must be 0 or 1. Got 1234."):
        QCVVExperiment._canonicalize_bitstring("1234", 4)


def test_canonicalize_probabilities() -> None:
    p: dict[str | int, float] = {3: 0.3, 0: 0.1, "10": 0.6}
    assert QCVVExperiment._canonicalize_probabilities(p, 2) == {
        "00": 0.1,
        "01": 0.0,
        "10": 0.6,
        "11": 0.3,
    }

    with pytest.raises(
        RuntimeError,
        match="Provided probabilities do not sum to 1.0. Got",  # Ignore exact value due to rounding
    ):
        p[0] = 0.0
        QCVVExperiment._canonicalize_probabilities(p, 2)



from __future__ import annotations

import copy

import cirq
import numpy as np
import numpy.typing as npt
import scipy.optimize as opt

import supermarq


class VQEProxy(supermarq.benchmark.Benchmark):
    """Proxy benchmark of a full VQE application that targets a single iteration
    of the whole variational optimization.

    The benchmark is parameterized by the number of qubits, n. For each value of
    n, we classically optimize the ansatz, sample 3 iterations near convergence,
    and use the sampled parameters to execute the corresponding circuits on the
    QPU. We take the measured energies from these experiments and average their
    values and compute a score based on how closely the experimental results are
    to the noiseless values.
    """

    def __init__(self, num_qubits: int, num_layers: int = 1) -> None:
        self.num_qubits = num_qubits
        self.num_layers = num_layers
        self.hamiltonian = self._gen_tfim_hamiltonian()
        self._params = self._gen_angles()

    def _gen_tfim_hamiltonian(self) -> list[tuple[str, int | tuple[int, int], int]]:
        r"""Generate an n-qubit Hamiltonian for a transverse-field Ising model (TFIM).

            $H = \sum_i^n(X_i) + \sum_i^n(Z_i Z_{i+1})$

        Example of a 6-qubit TFIM Hamiltonian:

            $H_6 = XIIIII + IXIIII + IIXIII + IIIXII + IIIIXI + IIIIIX + ZZIIII
                  + IZZIII + IIZZII + IIIZZI + IIIIZZ + ZIIIIZ$
        """
        hamiltonian: list[tuple[str, int | tuple[int, int], int]] = []
        for i in range(self.num_qubits):
            hamiltonian.append(("X", i, 1))  # [Pauli type, qubit idx, weight]
        for i in range(self.num_qubits - 1):
            hamiltonian.append(("ZZ", (i, i + 1), 1))
        hamiltonian.append(("ZZ", (self.num_qubits - 1, 0), 1))
        return hamiltonian

    def _gen_ansatz(self, params: npt.NDArray[np.float64]) -> list[cirq.Circuit]:
        qubits = cirq.LineQubit.range(self.num_qubits)
        z_circuit = cirq.Circuit()

        param_counter = 0
        for _ in range(self.num_layers):
            # Ry rotation block
            for i in range(self.num_qubits):
                z_circuit.append(cirq.Ry(rads=2 * params[param_counter])(qubits[i]))
                param_counter += 1
            # Rz rotation block
            for i in range(self.num_qubits):
                z_circuit.append(cirq.Rz(rads=2 * params[param_counter])(qubits[i]))
                param_counter += 1
            # Entanglement block
            for i in range(self.num_qubits - 1):
                z_circuit.append(cirq.CX(qubits[i], qubits[i + 1]))
            # Ry rotation block
            for i in range(self.num_qubits):
                z_circuit.append(cirq.Ry(rads=2 * params[param_counter])(qubits[i]))
                param_counter += 1
            # Rz rotation block
            for i in range(self.num_qubits):
                z_circuit.append(cirq.Rz(rads=2 * params[param_counter])(qubits[i]))
                param_counter += 1

        x_circuit = copy.deepcopy(z_circuit)
        x_circuit.append(cirq.H(q) for q in qubits)

        # Measure all qubits
        z_circuit.append(cirq.measure(*qubits))
        x_circuit.append(cirq.measure(*qubits))

        return [z_circuit, x_circuit]

    def _parity_ones(self, bitstr: str) -> int:
        one_count = 0
        for i in bitstr:
            if i == "1":
                one_count += 1
        return one_count % 2

    def _calc(self, bit_list: list[str], bitstr: str, probs: dict[str, float]) -> float:
        energy = 0.0
        for item in bit_list:
            if self._parity_ones(item) == 0:
                energy += probs.get(bitstr, 0)
            else:
                energy -= probs.get(bitstr, 0)
        return energy

    def _get_expectation_value_from_probs(
        self, probs_z: dict[str, float], probs_x: dict[str, float]
    ) -> float:
        avg_energy = 0.0

        # Find the contribution to the energy from the X-terms: \sum_i{X_i}
        for bitstr in probs_x.keys():
            bit_list_x = [bitstr[i] for i in range(len(bitstr))]
            avg_energy += self._calc(bit_list_x, bitstr, probs_x)

        # Find the contribution to the energy from the Z-terms: \sum_i{Z_i Z_{i+1}}
        for bitstr in probs_z.keys():
            # fmt: off
            bit_list_z = [bitstr[(i - 1): (i + 1)] for i in range(1, len(bitstr))]
            # fmt: on
            bit_list_z.append(bitstr[0] + bitstr[-1])  # Add the wrap-around term manually
            avg_energy += self._calc(bit_list_z, bitstr, probs_z)

        return avg_energy

    def _get_opt_angles(self) -> tuple[npt.NDArray[np.float64], float]:
        def f(params: npt.NDArray[np.float64]) -> float:
            """The objective function to minimize.

            Args:
                params: The parameters at which to evaluate the objective.

            Returns:
                Evaluation of objective given parameters.
            """
            z_circuit, x_circuit = self._gen_ansatz(params)
            z_probs = supermarq.simulation.get_ideal_counts(z_circuit)
            x_probs = supermarq.simulation.get_ideal_counts(x_circuit)
            energy = self._get_expectation_value_from_probs(z_probs, x_probs)

            return -energy  # because we are minimizing instead of maximizing

        init_params = [
            np.random.uniform() * 2 * np.pi for _ in range(self.num_layers * 4 * self.num_qubits)
        ]
        out = opt.minimize(f, init_params, method="COBYLA")

        return out["x"], out["fun"]

    def _gen_angles(self) -> npt.NDArray[np.float64]:
        """Classically simulate the variational optimization and return
        the final parameters.
        """
        params, _ = self._get_opt_angles()
        return params

    def circuit(self) -> list[cirq.Circuit]:
        """Construct a parameterized ansatz.

        The counts obtained from evaluating these two circuits should be passed to `score` in the
        same order they are returned here.

        Returns:
            A list of circuits for the VQE benchmark: the ansatz measured in the Z basis, and the
            ansatz measured in the X basis.
        """
        return self._gen_ansatz(self._params)

    def score(self, counts: list[dict[str, float]]) -> float:
        """Compare the average energy measured by the experiments to the ideal value.

        The ideal value is obtained via noiseless simulation. In principle the ideal value can be
        obtained through efficient classical means since the 1D TFIM is analytically solvable.

        Args:
            counts: A dictionary containing the measurement counts from circuit execution.

        Returns:
            The VQE proxy benchmark score.
        """
        counts_z, counts_x = counts
        shots_z = sum(counts_z.values())
        probs_z = {bitstr: count / shots_z for bitstr, count in counts_z.items()}
        shots_x = sum(counts_x.values())
        probs_x = {bitstr: count / shots_x for bitstr, count in counts_x.items()}
        experimental_expectation = self._get_expectation_value_from_probs(probs_z, probs_x)

        circuit_z, circuit_x = self.circuit()
        ideal_expectation = self._get_expectation_value_from_probs(
            supermarq.simulation.get_ideal_counts(circuit_z),
            supermarq.simulation.get_ideal_counts(circuit_x),
        )

        return float(
            1.0 - abs(ideal_expectation - experimental_expectation) / abs(2 * ideal_expectation)
        )



from __future__ import annotations

import cirq
import networkx as nx
import numpy as np
import qiskit


def cirq_to_qiskit(circuit: cirq.Circuit) -> qiskit.QuantumCircuit:
    """Convert a circuit from cirq to qiskit.

    Args:
        circuit: A `cirq.Circuit` to be converted.

    Returns:
        An equivalent `qiskit.QuantumCircuit`.
    """
    qasm = cirq.circuits.QasmOutput(circuit, tuple(sorted(circuit.all_qubits())))
    return qiskit.circuit.QuantumCircuit().from_qasm_str(str(qasm))


def compute_communication_with_qiskit(circuit: qiskit.QuantumCircuit) -> float:
    """Compute the program communication of the given quantum circuit.

    Program communication = circuit's average qubit degree / degree of a complete graph.

    Args:
        circuit: A quantum circuit.

    Returns:
        The value of the communication feature for this circuit.
    """
    num_qubits = circuit.num_qubits
    dag = qiskit.converters.circuit_to_dag(circuit)
    dag.remove_all_ops_named("barrier")

    graph = nx.Graph()
    for op in dag.two_qubit_ops():
        q1, q2 = op.qargs
        graph.add_edge(circuit.find_bit(q1).index, circuit.find_bit(q2).index)

    degree_sum = sum([graph.degree(n) for n in graph.nodes])

    return degree_sum / (num_qubits * (num_qubits - 1))


def compute_liveness_with_qiskit(circuit: qiskit.QuantumCircuit) -> float:
    """Compute the liveness of the given quantum circuit.

    Liveness feature = sum of all entries in the liveness matrix / (num_qubits * depth).

    Args:
        circuit: A quantum circuit.

    Returns:
        The value of the liveness feature for this circuit.
    """

    num_qubits = circuit.num_qubits
    dag = qiskit.converters.circuit_to_dag(circuit)
    dag.remove_all_ops_named("barrier")

    activity_matrix = np.zeros((num_qubits, dag.depth()))

    for i, layer in enumerate(dag.layers()):
        for op in layer["partition"]:
            for qubit in op:
                activity_matrix[circuit.find_bit(qubit).index, i] = 1

    return np.sum(activity_matrix) / (num_qubits * dag.depth())


def compute_parallelism_with_qiskit(circuit: qiskit.QuantumCircuit) -> float:
    """Compute the parallelism of the given quantum circuit.

    Parallelism feature = max((((# of gates / depth) - 1) /(# of qubits - 1)), 0).

    Args:
        circuit: A quantum circuit.

    Returns:
        The value of the parallelism feature for this circuit.
    """
    dag = qiskit.converters.circuit_to_dag(circuit)
    dag.remove_all_ops_named("barrier")
    if circuit.num_qubits <= 1:
        return 0
    depth = dag.depth()
    if depth == 0:
        return 0
    return max(((len(dag.gate_nodes()) / depth) - 1) / (circuit.num_qubits - 1), 0)


def compute_measurement_with_qiskit(circuit: qiskit.QuantumCircuit) -> float:
    """Compute the measurement feature of the given quantum circuit.

    Measurement feature = # of layers of mid-circuit measurement / circuit depth.

    Args:
        circuit: A quantum circuit.

    Returns:
        The value of the measurement feature for this circuit.
    """
    circuit.remove_final_measurements()
    dag = qiskit.converters.circuit_to_dag(circuit)
    dag.remove_all_ops_named("barrier")

    reset_moments = 0
    gate_depth = dag.depth()

    for layer in dag.layers():
        reset_present = False
        for op in layer["graph"].op_nodes():
            if op.name == "reset":
                reset_present = True
        if reset_present:
            reset_moments += 1

    return reset_moments / gate_depth


def compute_entanglement_with_qiskit(circuit: qiskit.QuantumCircuit) -> float:
    """Compute the entanglement-ratio of the given quantum circuit.

    Entanglement-ratio = ratio between # of 2-qubit gates and total number of gates in the
    circuit.

    Args:
        circuit: A quantum circuit.

    Returns:
        The value of the entanglement feature for this circuit.
    """
    dag = qiskit.converters.circuit_to_dag(circuit)
    dag.remove_all_ops_named("barrier")

    return len(dag.two_qubit_ops()) / len(dag.gate_nodes())


def compute_depth_with_qiskit(circuit: qiskit.QuantumCircuit) -> float:
    """Compute the critical depth of the given quantum circuit.

    Critical depth = # of 2-qubit gates along the critical path / total # of 2-qubit gates.

    Args:
        circuit: A quantum circuit.

    Returns:
        The value of the depth feature for this circuit.
    """
    dag = qiskit.converters.circuit_to_dag(circuit)
    dag.remove_all_ops_named("barrier")
    longest_paths = dag.count_ops_longest_path()
    n_ed = sum(
        [
            longest_paths[name]
            for name in {op.name for op in dag.two_qubit_ops()}
            if name in longest_paths
        ]
    )
    n_e = len(dag.two_qubit_ops())

    if n_ed == 0:
        return 0

    return n_ed / n_e



# Copyright 2021 The Cirq Developers
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tooling for cross entropy benchmark experiments.
"""
from __future__ import annotations

import itertools
from collections.abc import Iterable, Sequence
from dataclasses import dataclass

import cirq
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy
import seaborn as sns
import tqdm.auto
import tqdm.contrib.itertools

from supermarq.qcvv.base_experiment import QCVVExperiment, QCVVResults, Sample


@dataclass
class XEBResults(QCVVResults):
    """Results from an XEB experiment."""

    _circuit_fidelities: pd.DataFrame | None = None
    _cycle_fidelity_estimate: float | None = None
    """Estimated cycle fidelity."""
    _cycle_fidelity_estimate_std: float | None = None
    """Standard deviation for the cycle fidelity estimate."""

    @property
    def cycle_fidelity_estimate(self) -> float:
        """Returns:
        Estimated cycle fidelity."""
        if self._cycle_fidelity_estimate is None:
            raise self._not_analyzed
        return self._cycle_fidelity_estimate

    @property
    def cycle_fidelity_estimate_std(self) -> float:
        """Returns:
        Standard deviation for the cycle fidelity estimate."""
        if self._cycle_fidelity_estimate_std is None:
            raise self._not_analyzed
        return self._cycle_fidelity_estimate_std

    def _analyze(self) -> None:
        """Analyse the results and calculate the estimated circuit fidelity.

        Args:
            plot_results (optional): Whether to generate the data plots. Defaults to True.

        Raises:
            RuntimeError: If there is no data stored.

        Returns:
           The final results from the experiment.
        """
        if self.data is None:
            raise RuntimeError("No data stored. Cannot perform analysis.")
        self.data["sum_p(x)p^(x)"] = pd.DataFrame(
            (
                self.data[["00", "01", "10", "11"]].values
                * self.data[["exact_00", "exact_01", "exact_10", "exact_11"]].values
            ).sum(axis=1),
            index=self.data.index,
        )
        self.data["sum_p(x)p(x)"] = pd.DataFrame(
            (
                self.data[["exact_00", "exact_01", "exact_10", "exact_11"]].values
                * self.data[["exact_00", "exact_01", "exact_10", "exact_11"]].values
            ).sum(axis=1),
            index=self.data.index,
        )
        # Fit a linear model for each cycle depth to estimate the circuit fidelity
        records = []
        for depth in set(self.data.cycle_depth):
            df = self.data[self.data.cycle_depth == depth]
            fit = scipy.stats.linregress(
                x=df["sum_p(x)p(x)"] - 1 / 2**self.num_qubits,
                y=df["sum_p(x)p^(x)"] - 1 / 2**self.num_qubits,
            )

            records.append(
                {
                    "cycle_depth": depth,
                    "circuit_fidelity_estimate": fit.slope,
                    "circuit_fidelity_estimate_std": fit.stderr,
                }
            )

        self._circuit_fidelities = pd.DataFrame(records)

        self._circuit_fidelities["log_fidelity_estimate"] = np.log(
            self._circuit_fidelities.circuit_fidelity_estimate
        )

        # Fit a linear model to the depth ~ log(fidelity) to approximate the cycle fidelity
        cycle_fit = scipy.stats.linregress(
            x=self._circuit_fidelities.cycle_depth,
            y=np.log(self._circuit_fidelities.circuit_fidelity_estimate),
        )
        self._cycle_fidelity_estimate = np.exp(cycle_fit.slope)
        self._cycle_fidelity_estimate_std = self.cycle_fidelity_estimate * cycle_fit.stderr

    def plot_results(self) -> None:
        """Plot the experiment data and the corresponding fits.

        Raises:
            RuntimeError: If there is no data stored.
        """
        if self.data is None:
            raise RuntimeError("No data stored. Cannot plot results.")

        if self._circuit_fidelities is None:
            raise RuntimeError(
                "No stored dataframe of circuit fidelities. Something has gone wrong."
            )

        plot_1 = sns.lmplot(
            data=self.data,
            x="sum_p(x)p(x)",
            y="sum_p(x)p^(x)",
            hue="cycle_depth",
            palette="dark:r",
            legend="full",
            ci=None,
        )
        sns.move_legend(plot_1, "center right")
        ax_1 = plot_1.axes.item()
        plot_1.tight_layout()
        ax_1.set_xlabel(r"$\sum p(x)^2$", fontsize=15)
        ax_1.set_ylabel(r"$\sum p(x) \hat{p}(x)$", fontsize=15)
        ax_1.set_title(r"Linear fit per cycle depth", fontsize=15)

        plot_2 = sns.lmplot(
            data=self._circuit_fidelities,
            x="cycle_depth",
            y="circuit_fidelity_estimate",
            hue="cycle_depth",
            palette="dark:r",
        )
        ax_2 = plot_2.axes.item()
        plot_2.tight_layout()
        ax_2.set_xlabel(r"Cycle depth", fontsize=15)
        ax_2.set_ylabel(r"Circuit fidelity", fontsize=15)
        ax_2.set_title(r"Exponential decay of circuit fidelity", fontsize=15)

        # Add fit line
        x = np.linspace(
            self._circuit_fidelities.cycle_depth.min(), self._circuit_fidelities.cycle_depth.max()
        )
        y = self.cycle_fidelity_estimate**x
        y_p = (self.cycle_fidelity_estimate + self.cycle_fidelity_estimate_std) ** x
        y_m = (self.cycle_fidelity_estimate - self.cycle_fidelity_estimate_std) ** x
        ax_2.plot(x, y, color="tab:red", linewidth=2)
        ax_2.fill_between(x, y_m, y_p, alpha=0.2, color="tab:red")

    def print_results(self) -> None:
        print(
            f"Estimated cycle fidelity: {self.cycle_fidelity_estimate:.5} "
            f"+/- {self.cycle_fidelity_estimate_std:.5}"
        )

    def plot_speckle(self) -> None:
        """Creates the speckle plot of the XEB data. See Fig. S18 of
        https://arxiv.org/abs/1910.11333 for an explanation of this plot.
        """
        df = self.data
        df2 = pd.melt(
            df,
            value_vars=["00", "01", "10", "11"],
            id_vars=["cycle_depth", "circuit_index"],
            var_name="bitstring",
        )
        fig, axs = plt.subplots(nrows=4, sharex=True)
        fig.subplots_adjust(hspace=0)
        for k, bitstring in enumerate(["00", "01", "10", "11"]):
            data = df2[df2["bitstring"] == bitstring].pivot(
                index="circuit_index", columns="cycle_depth", values="value"
            )
            cmap = mpl.colormaps["rocket"]
            norm = mpl.colors.Normalize(0, 1)  # or vmin, vmax
            sns.heatmap(data, vmin=0, vmax=1, ax=axs[k], cbar=False, cmap=cmap)
            axs[k].set_ylabel("")
            axs[k].set_xlabel("")
            axs[k].set_yticks([0, 15])
            axs[k].set_yticklabels([0, 15])
            plt.text(
                0.99,
                0.90,
                f"P({bitstring})",
                ha="right",
                va="top",
                transform=axs[k].transAxes,
                color="white",
            )
            if k != 0:
                axs[k].axhline(y=0, linewidth=1.5, color="white", linestyle="--")
        fig.supxlabel("Cycle depth")
        fig.supylabel("Circuit Instance")
        fig.colorbar(
            mpl.cm.ScalarMappable(norm, cmap), ax=axs, orientation="vertical", label="Probability"
        )


class XEB(QCVVExperiment[XEBResults]):
    r"""Cross-entropy benchmarking (XEB) experiment.

    The XEB experiment can be used to estimate the combined fidelity of a repeating
    cycle of gates. In our case, where we restrict ourselves to two qubits, we use
    cycles made up of two randomly selected single qubit phased XZ gates and a constant
    two qubit gate. This is illustrated as follows:

    For each randomly generated circuit, with a given number of cycle, we compare the
    simulated state probabilities, :math:`p(x)` with those achieved by running the circuit
    on a given target, :math:`\hat{p}(x)`. The fidelity of a circuit containing :math:`d`
    cycles, :math:`f_d` can then be estimated as

    .. math::

        \sum_{x \in \{0, 1\}^n} p(x) \hat{p}(x) - \frac{1}{2^n} =
        f_d \left(\sum_{x \in \{0, 1\}^n} p(x)^2 -  \frac{1}{2^n}\right)

    We can therefore fit a linear model to estimate the value of :math:`f_d`. We the estimate
    the fidelity of the cycle, :math:`f_{\mathrm{cycle}}` as

    .. math::

        f_d = A(f_{cycle})^d

    Thus fitting another linear model to :math:`\log(f_d) \sim d` provides us with an estimate
    of the cycle fidelity.

    For more details see: https://www.nature.com/articles/s41586-019-1666-5
    """

    def __init__(
        self,
        num_circuits: int,
        cycle_depths: Iterable[int],
        single_qubit_gate_set: list[cirq.Gate] | None = None,
        two_qubit_gate: cirq.Gate | None = cirq.CZ,
        *,
        random_seed: int | np.random.Generator | None = None,
    ) -> None:
        """Initializes a cross-entropy benchmarking experiment.

        Args:
            num_circuits: Number of circuits to sample.
            cycle_depths: The cycle depths to sample.
            single_qubit_gate_set: Optional list of single qubit gates to randomly sample from when
                generating random circuits. If not provided defaults to phased XZ gates with 1/4 pi
                intervals.
            two_qubit_gate: The two qubit gate to interleave between the single qubit gates. If None
                then no two qubit gate is used. Defaults to control-Z gate.
            random_seed: An optional seed to use for randomization.
        """
        self.two_qubit_gate: cirq.Gate | None = two_qubit_gate
        """The two qubit gate to use for interleaving."""

        self.single_qubit_gate_set: list[cirq.Gate]
        """The single qubit gates to randomly sample from"""

        if single_qubit_gate_set is None:
            gate_exponents = np.linspace(start=0, stop=7 / 4, num=8)
            self.single_qubit_gate_set = [
                cirq.PhasedXZGate(
                    z_exponent=z,  # 1) Choose an axis in the xy-plane, zπ from the +x-axis.
                    x_exponent=0.5,  # 2) Rotate about the axis in 1) by a fixed π/2.
                    axis_phase_exponent=a,  # 3) Rotate about the +z-axis by aπ (a final phasing).
                )
                for a, z in itertools.product(
                    gate_exponents, repeat=2
                )  # enumerates every possible (a, z)
            ]
        else:
            self.single_qubit_gate_set = single_qubit_gate_set

        super().__init__(
            num_qubits=2,
            num_circuits=num_circuits,
            cycle_depths=cycle_depths,
            random_seed=random_seed,
            results_cls=XEBResults,
        )

    ###################
    # Private Methods #
    ###################
    def _build_circuits(
        self,
        num_circuits: int,
        cycle_depths: Iterable[int],
    ) -> Sequence[Sample]:
        """Build a list of random circuits to perform the XEB experiment with.

        Args:
            num_circuits: Number of circuits to generate.
            cycle_depths: An iterable of the different numbers of cycles to include in each circuit.

        Returns:
            The list of experiment samples.
        """
        random_circuits = []
        for k, depth in tqdm.contrib.itertools.product(
            range(num_circuits), cycle_depths, desc="Building circuits"
        ):
            num_single_qubit_gate_layers = depth + int(self.two_qubit_gate is not None)
            chosen_single_qubit_gates = self._rng.choice(
                np.asarray(self.single_qubit_gate_set),
                size=(num_single_qubit_gate_layers, self.num_qubits),
            )

            circuit = cirq.Circuit(
                gate.on(qubit)
                for gates_in_layer in chosen_single_qubit_gates
                for gate, qubit in zip(gates_in_layer, self.qubits)
            )

            if self.two_qubit_gate is not None:
                circuit = self._interleave_op(circuit, self.two_qubit_gate(*self.qubits))

            analytic_final_state = cirq.final_state_vector(
                circuit, qubit_order=sorted(circuit.all_qubits())
            )
            analytic_probabilities = {
                "exact_" + format(idx, f"0{self.num_qubits}b"): np.abs(state) ** 2
                for idx, state in enumerate(analytic_final_state)
            }

            random_circuits.append(
                Sample(
                    circuit=circuit + cirq.measure(sorted(circuit.all_qubits())),
                    data={
                        "circuit_depth": len(circuit),
                        "cycle_depth": depth,
                        "two_qubit_gate": str(self.two_qubit_gate),
                        **analytic_probabilities,
                    },
                    circuit_index=k,
                )
            )

        return random_circuits



from __future__ import annotations

import os
from collections.abc import Generator
from unittest import mock

import pytest


@pytest.fixture(scope="session", autouse=True)
def _patch_tqdm() -> Generator[None, None, None]:
    """Disable progress bars during tests."""
    with mock.patch.dict(os.environ, {"TQDM_DISABLE": "1"}):
        yield



from __future__ import annotations

from collections.abc import Mapping

import cirq
import numpy as np
import qiskit

import supermarq
from supermarq.benchmark import Benchmark


class HamiltonianSimulation(Benchmark):
    """Quantum benchmark focused on the ability to simulate 1D
    Transverse Field Ising Models (TFIM) of variable length.

    Device performance is based on how closely the experimentally obtained
    average magnetization (along the Z-axis) matches the noiseless value.
    Since the 1D TFIM is efficiently simulatable with classical algorithms,
    computing the noiseless average magnetization remains scalable over a large
    range of benchmark sizes.
    """

    def __init__(self, num_qubits: int, time_step: int = 1, total_time: int = 1) -> None:
        """Initializes a `HamiltonianSimulation`.

        Args:
            num_qubits: Size of the TFIM chain, equivalent to the number of qubits.
            time_step: Size of the timestep in attoseconds. Defaults to 1.
            total_time: Total simulation time of the TFIM chain in attoseconds. Defaults to 1.
        """
        self.num_qubits = num_qubits
        self.time_step = time_step
        self.total_time = total_time

    def circuit(self) -> cirq.Circuit:
        """Generates a circuit to simulate the evolution of an n-qubit TFIM
        chain under the Hamiltonian:

        H(t) = - Jz * sum_{i=1}^{n-1}(sigma_{z}^{i} * sigma_{z}^{i+1})
               - e_ph * cos(w_ph * t) * sum_{i=1}^{n}(sigma_{x}^{i})

        where,
            w_ph: frequency of E" phonon in MoSe2.
            e_ph: strength of electron-phonon coupling.

        Returns:
            The circuit for Hamiltonian simulation.
        """
        hbar = 0.658212  # eV*fs
        jz = (
            hbar * np.pi / 4
        )  # eV, coupling coeff; Jz<0 is antiferromagnetic, Jz>0 is ferromagnetic
        freq = 0.0048  # 1/fs, frequency of MoSe2 phonon

        w_ph = 2 * np.pi * freq
        e_ph = 3 * np.pi * hbar / (8 * np.cos(np.pi * freq))

        qubits = cirq.LineQubit.range(self.num_qubits)
        circuit = cirq.Circuit()

        # Build up the circuit over total_time / time_step propagation steps
        for step in range(int(self.total_time / self.time_step)):
            # Simulate the Hamiltonian term-by-term
            t = (step + 0.5) * self.time_step

            # Single qubit terms
            psi = -2.0 * e_ph * np.cos(w_ph * t) * self.time_step / hbar
            for qubit in qubits:
                circuit.append(cirq.H(qubit))
                circuit.append(cirq.rz(psi)(qubit))
                circuit.append(cirq.H(qubit))

            # Coupling terms
            psi2 = -2.0 * jz * self.time_step / hbar
            for i in range(len(qubits) - 1):
                circuit.append(cirq.CNOT(qubits[i], qubits[i + 1]))
                circuit.append(cirq.rz(psi2)(qubits[i + 1]))
                circuit.append(cirq.CNOT(qubits[i], qubits[i + 1]))

        # End the circuit with measurements of every qubit in the Z-basis
        circuit.append(cirq.measure(*qubits))

        return circuit

    def qiskit_circuit(self) -> qiskit.QuantumCircuit:
        """Generates a circuit to simulate the evolution of an n-qubit TFIM
        chain under the Hamiltonian:

        H(t) = - Jz * sum_{i=1}^{n-1}(sigma_{z}^{i} * sigma_{z}^{i+1})
               - e_ph * cos(w_ph * t) * sum_{i=1}^{n}(sigma_{x}^{i})

        where,
            w_ph: frequency of E" phonon in MoSe2.
            e_ph: strength of electron-phonon coupling.

        Returns:
            The circuit for Hamiltonian simulation.
        """
        hbar = 0.658212  # eV*fs
        jz = (
            hbar * np.pi / 4
        )  # eV, coupling coeff; Jz<0 is antiferromagnetic, Jz>0 is ferromagnetic
        freq = 0.0048  # 1/fs, frequency of MoSe2 phonon

        w_ph = 2 * np.pi * freq
        e_ph = 3 * np.pi * hbar / (8 * np.cos(np.pi * freq))

        circuit = qiskit.QuantumCircuit(self.num_qubits, self.num_qubits)

        # Build up the circuit over total_time / time_step propagation steps
        for step in range(int(self.total_time / self.time_step)):
            # Simulate the Hamiltonian term-by-term
            t = (step + 0.5) * self.time_step

            # Single qubit terms
            psi = -2.0 * e_ph * np.cos(w_ph * t) * self.time_step / hbar
            for i in range(self.num_qubits):
                circuit.h(i)
                circuit.rz(psi, i)
                circuit.h(i)

            # Coupling terms
            psi2 = -2.0 * jz * self.time_step / hbar
            for i in range(self.num_qubits - 1):
                circuit.cx(i, i + 1)
                circuit.rz(psi2, i + 1)
                circuit.cx(i, i + 1)

        # End the circuit with measurements of every qubit in the Z-basis
        for i in range(self.num_qubits):
            circuit.measure(i, i)

        return circuit

    def _average_magnetization(self, result: Mapping[str, float], shots: float) -> float:
        mag = 0.0
        for spin_str, count in result.items():
            spin_int = [1 - 2 * int(s) for s in spin_str]
            mag += (
                sum(spin_int) / len(spin_int)
            ) * count  # <Z> weighted by number of times we saw this bitstring
        average_mag = mag / shots  # normalize by the total number of shots
        return average_mag

    def score(self, counts: Mapping[str, float]) -> float:
        """Compute the average magnetization of the TFIM chain along the Z-axis
        for the experimental results and via noiseless simulation.

        Args:
            counts: Dictionary of the experimental results. The keys are bitstrings
                represented the measured qubit state, and the values are the number
                of times that state of observed.

        Returns:
            The Hamiltonian simulation benchmark score.
        """
        ideal_counts = supermarq.simulation.get_ideal_counts(self.circuit())

        total_shots = sum(counts.values())

        mag_ideal = self._average_magnetization(ideal_counts, 1)
        mag_experimental = self._average_magnetization(counts, total_shots)

        return 1 - abs(mag_ideal - mag_experimental) / 2
